// cameraops.h: hw-dependent camera control functions
// $Revision: 1.5 $
// $Date: 2003/06/20 23:43:52 $

#ifndef _CAMERAOPS_H

#define _CAMERAOPS_H

// camera capabilities structure
struct cam_info {
	unsigned active_pixels; // number of active pixels per line
	unsigned active_lines; // number of active frames
	unsigned dummy_pix_start; // number of inactive pixels at the start of each line
	unsigned dummy_pix_end; // number of inactive pixels at the end of each line
	unsigned dummy_lines_start; // number of dummy lines at the beginning of frame
	unsigned dummy_lines_end; // number of dummy lines at the beginning of frame
	unsigned pix_x_size; // physical x size of unbinned pixel (0.1u units)
	unsigned pix_y_size; // physical y size of unbinned pixel (0.1u units)
	unsigned min_exp; // minimum exposure (ms)
	unsigned max_exp; // maximum exposure (ms)
	int pipeline_mode; // '1' when the camera is capable of exposing a new frame while reading
	char camera_name[FRAME_NAME_SIZE];
};

// values describing the geometry of a frame from a ccd chip
struct ccd_geometry {
	unsigned w; /* total size of the frame */
	unsigned h; 
	int x_skip; /* number of pixels skipped at the beginning of each line 
		     * (ref: first active pixel) */
	int y_skip; /* number of lines skipped at the beginning of each frame */
};

// structure describing a ccd exposure 
struct exp_data {
	int datavalid;
	unsigned bin_x; // number of pixels binned in the x direction
	unsigned bin_y; // number of pixels binned in the y direction
	int fast_flush; // set to '1' when a fast flush is required, usually for focusing
	unsigned exp_time; // exposure time (ms)
	unsigned gain; // the gain setting for the camera
	unsigned read_spd; // the readout speed (e.g. adc conversions/value)
	double scale; // the scale of the data (in electrons/ADU)
	double bias; // the average bias of the data
	double rdnoise; // the readout noise
	double flat_noise; // relative noise of flatfield
	int temperature; // sensor temperature
	unsigned cooling_power; // actual cooling power 
	struct timeval exp_start; // exposure start time
};

/* camera states */
#define CAM_INVALID 	0 // camera is in an invalid state (not initialised or hung up)
#define CAM_IDLE 	1 // camera is idle
#define CAM_EXP 	2 // camera is integrating
#define CAM_EXPEND	3 // camera is at the end of exposure and can begin reading
#define CAM_READ 	4 // camera is ready to read / reading out an image
#define CAM_READ_END 	5 // camera is at end of read, waiting for the data fifo to be 
// emptied

#define CAM_STAT_MASK 0x0f

// magic numbers for cpx3m cameras connected to head1
#define CPX3M_MAIN_MAGIC1 0xcc33cc11
#define CPX3M_GUIDE_MAGIC1 0x11223311

// driver private data declaration
struct cpx3m_priv {
//	unsigned state;
	int linesread; // number of lines read so far in the current op
	int got_shutter; // flag showing that we hold the shutter lock
	struct timeval sns_timer; // timer for sensor ops
	struct timeval exp_timer; // exposure timer
};

// the camera structure contains actual settings for a camera
// a different structure exists for each camera attached
struct camera {
	unsigned magic; // an unique number identifying the camera
	int state; // camera state (see defines below)
	struct cam_info info;
	struct ccd_geometry geom; // the default geometry for this camera
	struct exp_data exp;
	unsigned exp_elapsed; // time that that has elapsed from the start of an exposure
	struct dbuf *db; /* data buffer */
	int lock;
};


// shutter control struct - private driver data
struct shutter {
	int lock; // lock flag
	int shtr_enabled; // flag showing that the shutter was actually open
	struct timeval open_time; // time the shutter was last open
	struct timeval close_time; // time the shutter eas last closed
};

// temperature control struct - private driver data
struct tempctrl {
	unsigned cooling_power_i; // I term of cooling controller
	int treg_p; // regulator coefficients (inverse)
	int treg_i;
	int temperature; // sensor temperature
	unsigned cooling_power; // actual cooling power 
	int temp_set; // temperature setpoint (0.1 degrees C units)
	struct timeval temp_timer; // timer for temp regulator
};

extern struct camera main_cam;
extern struct camera track_cam;

// from cameraops.c

extern int cameras_init(void);
extern int reset_camera(struct camera *cam);
extern int temp_reg(struct camera *cam);
extern int exp_start(struct camera *cam, int no_flush, int shtren);
extern int read_img(struct camera *cam, void *buf, int timeout);
extern int close_shutter(struct camera *cam);
extern int open_shutter(struct camera *cam);
extern int cam_machine(struct camera *cam);
int set_exp(struct camera *cam, unsigned exptime);
void cam_set_buffer(struct camera * cam, struct dbuf *db);

#endif
