/*******************************************************************************
  Copyright(c) 2000 - 2003 Radu Corlan. All rights reserved.
  
  This program is free software; you can redistribute it and/or modify it 
  under the terms of the GNU General Public License as published by the Free 
  Software Foundation; either version 2 of the License, or (at your option) 
  any later version.
  
  This program is distributed in the hope that it will be useful, but WITHOUT 
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
  more details.
  
  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc., 59 
  Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  The full GNU General Public License is included in this distribution in the
  file called LICENSE.
  
  Contact Information: radu@corlan.net
*******************************************************************************/


#define _GNU_SOURCE
#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include <string.h>
#include <ctype.h>

#include "cpxctrl.h"
#include "params.h"

struct param ptable[PAR_TABLE_SIZE];

/* functions used for parameter table initialisation */

/* find an attachemnt node (last of parent's children)
 * return -1 if no suitable point is found, or the node.
 * a link to the new item is added in the found node
 */
static int add_link(GcxPar parent, GcxPar p)
{
	int point;
	if (parent == PAR_NULL) {
		point = PAR_FIRST;
		while (PAR(point)->next != PAR_NULL)
			point = PAR(point)->next;
		PAR(point)->next = p;
		return point;
	}
	if (PAR_TYPE(parent) != PAR_TREE) {
		err_printf("add_par: parent %d is not a tree!\n", parent);
		return -1;
	}
	if (PAR(parent)->child == PAR_NULL) {
		PAR(parent)->child = p;
		return parent;
	}
	point = PAR(parent)->child;
	while (PAR(point)->next != PAR_NULL)
		point = PAR(point)->next;
	PAR(point)->next = p;
	return point;
}

static void add_par_tree(GcxPar p, GcxPar parent, char *name, char *comment)
{
	int point;

	if (p >= PAR_TABLE_SIZE || parent >= PAR_TABLE_SIZE) {
		err_printf("add_par: bad par[%d] index \n", p);
		return;
	}
	point = add_link(parent, p);
	if (point <= 0) {
		err_printf("add_par: cannot find link point for [%d]\n", p);
		return;
	}
	PAR(p)->parent = (parent);
	PAR(p)->child = PAR_NULL;
	PAR(p)->next = PAR_NULL;
	PAR(p)->flags = PAR_TREE;
	PAR(p)->name = name;
	PAR(p)->comment = comment;
}

static void add_par_int(GcxPar p, GcxPar parent, int format, char *name, char *comment, int v)
{
	int point;

	if (p >= PAR_TABLE_SIZE || parent >= PAR_TABLE_SIZE) {
		err_printf("add_par: bad par index\n");
		return;
	}
	point = add_link(parent, p);
	if (point <= 0) {
		err_printf("add_par: cannot find link point for [%d]\n", p);
		return;
	}
	PAR(p)->parent = (parent);
	PAR(p)->child = (PAR_NULL);
	PAR(p)->next = PAR_NULL;
	PAR(p)->flags = PAR_INTEGER | format;
	PAR(p)->name = name;
	PAR(p)->comment = comment;
	P_INT(p) = v;
	PAR(p)->defval.i = v;
}

static void add_par_double(GcxPar p, GcxPar parent, int format, char *name, char *comment, 
			   double v)
{
	int point;

	if (p >= PAR_TABLE_SIZE || parent >= PAR_TABLE_SIZE) {
		err_printf("add_par: bad par index\n");
		return;
	}
	point = add_link(parent, p);
	if (point <= 0) {
		err_printf("add_par: cannot find link point for [%d]\n", p);
		return;
	}
	PAR(p)->parent = (parent);
	PAR(p)->child = (PAR_NULL);
	PAR(p)->next = PAR_NULL;
	PAR(p)->flags = PAR_DOUBLE | format;
	PAR(p)->name = name;
	PAR(p)->comment = comment;
	P_DBL(p) = v;
	PAR(p)->defval.d = v;
}

static void add_par_string(GcxPar p, GcxPar parent, int format, char *name, 
			   char *comment, char *val)
{
	int point;

	if (p >= PAR_TABLE_SIZE || parent >= PAR_TABLE_SIZE) {
		err_printf("add_par: bad par index\n");
		return;
	}
	point = add_link(parent, p);
	if (point <= 0) {
		err_printf("add_par: cannot find link point for [%d]\n", p);
		return;
	}
	PAR(p)->parent = (parent);
	PAR(p)->child = (PAR_NULL);
	PAR(p)->next = PAR_NULL;
	PAR(p)->flags = PAR_STRING | format;
	PAR(p)->name = name;
	PAR(p)->comment = comment;
	if (P_STR(p) != NULL) {
		free(P_STR(p));
		P_STR(p) = NULL;
	}
	if (PAR(p)->defval.s != NULL) {
		free(PAR(p)->defval.s);
		PAR(p)->defval.s = NULL;
	}
	P_STR(p) = malloc(strlen(val)+1);
	if (P_STR(p) != NULL)
		strcpy(P_STR(p), val);
	PAR(p)->defval.s = malloc(strlen(val)+1);
	if (PAR(p)->defval.s != NULL)
		strcpy(PAR(p)->defval.s, val);
}

/*
 * make an int parameter show as a choices list
 * choices must be valid for the lifetime of the
 * program; static constants are good*/
static void set_par_choices(GcxPar p, char ** choices)
{
	int i = 0;
	if (PAR_TYPE(p) != PAR_INTEGER) {
		err_printf("cannot add choices to a non-int par\n");
		return;
	}
	PAR(p)->choices = choices;
	PAR(p)->flags &= ~PAR_FORMAT_FM;
	PAR(p)->flags |= FMT_OPTION;
	while (choices[i] != NULL) 
		i++;
	clamp_int(&(P_INT(p)), 0, i);
}



/* the initialisation proper */


void add_camera_pars(GcxPar cam)
{
	add_par_tree(M_CAM_INFO, cam,
		     "info", "Camera capabilities");
	add_par_tree(M_CAM_EXP, cam,
		     "exp", "Exposure settings");
	add_par_tree(M_CAM_STAT, cam,
		     "stat", "Current camera status");

/* info items */
	add_par_string(M_CAM_NAME, M_CAM_INFO, 0, "name", 
		       "Camera name", "cpx3m");
	add_par_int(M_ACTIVE_PIX, M_CAM_INFO, 0, "active_pix", 
		    "Active pixels per line", 2184);
	add_par_int(M_ACTIVE_LINES, M_CAM_INFO, 0, "active_lines", 
		    "Active lines", 1472);
	add_par_int(M_DUMMY_PIX_START, M_CAM_INFO, 0, "dummy_pix_start", 
		    "Dummy pixels at start of line", 46);
	add_par_int(M_DUMMY_PIX_END, M_CAM_INFO, 0, "dummy_pix_end", 
		    "Dummy pixels at end of line", 37);
	add_par_int(M_DUMMY_LINES_START, M_CAM_INFO, 0, "dummy_lines_start", 
		    "Dummy lines at start of frame", 34);
	add_par_int(M_DUMMY_LINES_END, M_CAM_INFO, 0, "dummy_lines_end", 
		    "Dummy lines at end of frame", 4);
	add_par_double(M_PIX_X_SIZE, M_CAM_INFO, 0, "pix_x_size",
		       "Pixel size in microns (x)", 6.8);
	add_par_double(M_PIX_Y_SIZE, M_CAM_INFO, 0, "pix_y_size",
		       "Pixel size in microns (y)", 6.8);
	add_par_double(M_MIN_EXP, M_CAM_INFO, 0, "min_exp",
		       "Min exposure time", 0.05);
	add_par_double(M_MAX_EXP, M_CAM_INFO, 0, "max_exp",
		       "Max exposure time", 3600.0);
	add_par_double(M_RDNOISE, M_CAM_INFO, 0, "rdnoise", 
		    "Readout noise (ADUs)", 7.0);
	add_par_double(M_SCALE, M_CAM_INFO, 0, "scale", 
		    "Scale (electrons / ADU)", 2.0);

/* exp items */
	add_par_int(M_WIDTH, M_CAM_EXP, PAR_WREN, "width", 
		    "Image width", 2184);
	add_par_int(M_HEIGHT, M_CAM_EXP, PAR_WREN, "height", 
		    "Image height", 1472);
	add_par_int(M_BIN_X, M_CAM_EXP, PAR_WREN, "bin_x", 
		    "Binning in the 'x' direction", 1);
	add_par_int(M_BIN_Y, M_CAM_EXP, PAR_WREN, "bin_y", 
		    "Binning in the 'y' direction", 1);
	add_par_int(M_SKIP_X, M_CAM_EXP, PAR_WREN, "skip_x", 
		    "X origin of active area", 0);
	add_par_int(M_SKIP_Y, M_CAM_EXP, PAR_WREN, "skip_y", 
		    "Y origin of active area", 0);
	add_par_double(M_EXPTIME, M_CAM_EXP, PAR_WREN, "exptime", 
		    "Integration time (seconds)", 1.0);
	add_par_double(M_JDATE, M_CAM_EXP, PREC_8, "jdate", 
		    "Julian date of start of exposure", 1.0);

/* stat items */
	add_par_int(M_STATUS, M_CAM_STAT, 0, "status", 
		    "Camera status", 0);
	add_par_double(M_EXP_LEFT, M_CAM_STAT, 0, "exp_left", 
		    "Exposure time left", 1.0);
	add_par_int(M_LINES_READ, M_CAM_STAT, 0, "lines_read", 
		    "Lines read so far", 0);
}

void init_ptable(void)
{

/* first the trees */
	add_par_tree(MAIN_CAM, PAR_NULL,
		     "main", "Main camera operating params");
//	add_par_tree(TRACK_CAM, PAR_NULL,
//		     "track", "Track camera operating params");
	add_par_tree(COOLER, PAR_NULL,
		     "cooler", "Cooling system params");

/* now leaves */
/* cooler */
	add_par_double(TEMP_SETPOINT, COOLER, PREC_1 | PAR_WREN, "temp_set", 
		       "Controller setpoint temperature", -35.0);
	add_par_double(TEMP_CURRENT, COOLER, PREC_1, "temp", 
		       "Current temperature", 0);
	add_par_double(COOLING_POWER, COOLER, PREC_2, "power", 
		       "Current temperature", 0.5);

	add_camera_pars(MAIN_CAM);
}

/*
 * return a pointer to a 1-char string describing the current
 * parameter status. The string is actually a constant
 */
char *status_string(GcxPar p)
{
	if (PAR(p)->flags & PAR_USER) {
		return "[ User Set ]";
	} else if (PAR(p)->flags & PAR_FROM_RC) {
		return "[ From File ]";
	} else {
		return "[ Default ]";
	}
}

/* print a item value for display) */
void make_value_string(GcxPar p, char *c)
{
	switch(PAR_TYPE(p)) {
	case PAR_INTEGER:
		sprintf(c, "%d", P_INT(p));
		break;
	case PAR_DOUBLE:
		switch(PAR(p)->flags & PAR_PREC_FM) {
		case PREC_1:
			sprintf(c, "%.1f", P_DBL(p));
			break;
		case PREC_2:
			sprintf(c, "%.2f", P_DBL(p));
			break;
		case PREC_4:
			sprintf(c, "%.4f", P_DBL(p));
			break;
		case PREC_8:
			sprintf(c, "%.8f", P_DBL(p));
			break;
		default:
			sprintf(c, "%.3f", P_DBL(p));
			break;
		}
		break;
	case PAR_STRING:
		sprintf(c, "%s", P_STR(p));
		break;
	default:
		err_printf("unknown param type %x\n", PAR(p)->flags);
	}
}

/* try to parse the string to a double according to the format
 * return 0 if successfull. 
 * We need to add stuff for formats other than the 'natural'
 */
int try_parse_double(GcxPar p, char *text)
{
	float v;
	int ret;
	ret = sscanf(text, "%f", &v);
	if (ret > 0) {
		P_DBL(p) = v;
		return 0;
	} else {
		return -1;
	}
}

/* try to parse the string to an int according to the format
 * return 0 if successfull. 
 * We need to add stuff for formats other than the 'natural'
 */
int try_parse_int(GcxPar p, char *text)
{
	int v;
	int ret;
	char ** c;

	if (PAR_FORMAT(p) == FMT_OPTION && PAR(p)->choices != NULL) {
		c = PAR(p)->choices;
		v = 0;
		while (*c != NULL) {
			if (!strcmp(text, *c)) {
				P_INT(p) = v;
				return 0;
			}
			v++;
			c++;
		}
		err_printf("try_parse_int: unknown choice %s\n", text);
		return -1;
	} 

	ret = sscanf(text, "%d", &v);
	if (ret > 0) {
		P_INT(p) = v;
		return 0;
	} else {
		return -1;
	}
}

/* change the parameter's value string using a copy of the supplied 
 * text */ 
void change_par_string(GcxPar p, char *text)
{
	if (P_STR(p) != NULL) {
		free(P_STR(p));
		P_STR(p) = NULL;
	}
	P_STR(p) = malloc(strlen(text)+1);
	if (P_STR(p) != NULL)
		strcpy(P_STR(p), text);
}

/* change the parameter's value string using a copy of the supplied 
 * text */ 
void change_par_default_string(GcxPar p, char *text)
{
	if (PAR(p)->defval.s != NULL) {
		free(PAR(p)->defval.s);
		PAR(p)->defval.s = NULL;
	}
	PAR(p)->defval.s = malloc(strlen(text)+1);
	if (PAR(p)->defval.s != NULL)
		strcpy(PAR(p)->defval.s, text);
}


/* parse the string according to the param's type and format
 * and update it's value. return 0 if the parse was successfull,
 * or a negative error
 */
int try_update_par_value(GcxPar p, char *text)
{
	if((PAR_FLAGS(p) & PAR_WREN) == 0) {
		err_printf("par %d is read-only\n", p);
		return -1;
	}
	switch(PAR_TYPE(p)) {
	case PAR_INTEGER:
		return try_parse_int(p, text);
	case PAR_DOUBLE:
		return try_parse_double(p, text);
	case PAR_STRING:
		change_par_string(p, text);
		return 0;
	default:
		err_printf("unknown param type %x\n", PAR(p)->flags);
		return -1;
	}
}

/* restore an item's default (recurses through children) */
/* the status of ancestor trees is not changed */
void par_item_restore_default(GcxPar p)
{
	if (PAR(p)->flags & (PAR_TREE)) { // recurse tree
		GcxPar pp;
		PAR(p)->flags &= ~PAR_USER;
		pp = PAR(p)->child;
		while(pp != PAR_NULL) {
			par_item_restore_default(pp);
			pp = PAR(pp)->next;
		}
	} else if (PAR(p)->flags & (PAR_USER)) {
			if (PAR_TYPE(p) == PAR_STRING) {
				change_par_string(p, PAR(p)->defval.s);
			} else {
				memcpy(&(PAR(p)->val), &(PAR(p)->defval), sizeof(union pval));
			}
			PAR(p)->flags &= ~PAR_USER; 
			PAR(p)->flags &= ~PAR_TO_SAVE;
	}
}

/* return the logic or of all children of a par subtree */
int or_child_type(GcxPar p, int type)
{
	if (PAR_TYPE(p) != PAR_TREE || PAR(p)->child == PAR_NULL) 
		return (type | PAR(p)->flags);
	else {
		type |= PAR(p)->flags;
		p = PAR(p)->child;
		while(p != PAR_NULL) {
			type |= or_child_type(p, type);
			p = PAR(p)->next;
		}
	}
	return type;
}

/* update the state of all ancestors to a node */
void update_ancestors_state(GcxPar p)
{
	p = PAR(p)->parent;
	while (p != PAR_NULL && PAR_TYPE(p) == PAR_TREE) {
		int type = 0;
		PAR(p)->flags &= ~(PAR_USER | PAR_FROM_RC);
		type = or_child_type(p, 0);
		PAR(p)->flags |= (type & (PAR_USER | PAR_FROM_RC));
		p = PAR(p)->parent;
	}
}
/* print indents*/
static void fprint_indent(FILE *fp, int level)
{
	int i;
	for (i=0; i< 2 * level; i++) 
		fputc(' ', fp);
}

/* print a leaf node */
static void fprint_par_leaf(FILE *fp, GcxPar p, int level)
{
	char buf[256];
	if (level == 0) {
		make_value_string(p, buf);
		if (PAR_TYPE(p) == PAR_STRING) {
			fprintf(fp, "'%s' ", buf);
		} else {
			fprintf(fp, "%s ", buf);
		}
		fprintf(fp, "\n");
		return;
	}
	fprint_indent(fp, level);
//	fprintf(fp, "# %s\n", PAR(p)->comment);
	if (PAR_TYPE(p) == PAR_INTEGER 
	    && PAR_FORMAT(p) == FMT_OPTION 
	    && PAR(p)->choices != NULL) {
/*		fprint_indent(fp, level);
		c = PAR(p)->choices;
		fprintf(fp, "# Choices: ");
		while (*c != NULL) {
			fprintf(fp, "%s", *c);
			c++;
			if (*c != NULL)
				fprintf(fp, ", ");
		}
		fputc('\n', fp);
		c = PAR(p)->choices;
		for (i = 0; i < P_INT(p) && *c != NULL; i++)
			c++;
		fprint_indent(fp, level);
		if ((PAR_FLAGS(p) & (PAR_FROM_RC | PAR_USER)) == 0) {
			fprintf(fp, "# ");
		}
		fprintf(fp, "%s %s\n\n", PAR(p)->name, *c);
*/	} else {
		make_value_string(p, buf);
/*		fprint_indent(fp, level);
		if ((PAR_FLAGS(p) & (PAR_FROM_RC | PAR_USER)) == 0) {
			fprintf(fp, "# ");
		}
*/
		if (PAR_TYPE(p) == PAR_STRING) {
			fprintf(fp, "%s '%s' ",PAR(p)->name, buf);
		} else {
			fprintf(fp, "%s %s ",PAR(p)->name, buf);
		}
	}
	fprintf(fp, "\n");
}

static void fprint_item(FILE *fp, GcxPar p, int level)
{
	int pp = p;
	if(PAR_TYPE(p) == PAR_TREE) {
//		sprintf(npath, "%s.%s", path, PAR(p)->name);
//		fprintf(fp, "# %s\n%s.\n\n", PAR(p)->comment, npath);
		if (level == 0) {
			fprintf(fp, "(\n");
		} else {
			fprint_indent(fp, level);
			fprintf(fp, "%s (\n", PAR(p)->name);
		}
		p = PAR(pp)->child;
/* output the leaves first */
		while (p != PAR_NULL) {
			if (PAR_TYPE(p) != PAR_TREE)
				fprint_item(fp, p, level + 1);
			p = PAR(p)->next;
		}
/* then the subtrees */
		p = PAR(pp)->child;
		while (p != PAR_NULL) {
			if (PAR_TYPE(p) == PAR_TREE)
				fprint_item(fp, p, level + 1);
			p = PAR(p)->next;
		}
//		fprintf(fp, "\n");
		fprint_indent(fp, level);
		fprintf(fp, ")\n");
	} else {
		fprint_par_leaf(fp, p, level);
	}
}




/* save params in a file 
 * if p is PAR_NULL, save all the params, otherwise just the 
 * given param and it's children
 * returns a negative error code, or 0 for success
 */
void fprint_params(FILE *fp, GcxPar p)
{
	if (p != PAR_NULL) {
		fprint_item(fp, p, 0);
	} else {
		p = PAR_FIRST;
		fprintf(fp, "(\n");
		while (p != PAR_NULL) {
			fprint_item(fp, p, 1);
			p = PAR(p)->next;
		}
		fprintf(fp, ")\n");
	}
}


#define buf_snprintf(obuf, last, format, args...) do {*obuf += snprintf(*obuf, last - *obuf, \
format, ## args); if (*obuf > last) *obuf = last;} while(0) 

/* print a leaf node */
static void snprint_par_leaf(char **obuf, char *last, GcxPar p, int level)
{
	char buf[256];
	if (level == 0) {
		make_value_string(p, buf);
		if (PAR_TYPE(p) == PAR_STRING) {
			buf_snprintf(obuf, last, "'%s' ", buf);
		} else {
			buf_snprintf(obuf, last, "%s ", buf);
		}
		return;
	}
	if (PAR_TYPE(p) == PAR_INTEGER 
	    && PAR_FORMAT(p) == FMT_OPTION 
	    && PAR(p)->choices != NULL) {
	} else {
		make_value_string(p, buf);
		if (PAR_TYPE(p) == PAR_STRING) {
			buf_snprintf(obuf, last, "%s '%s' ",PAR(p)->name, buf);
		} else {
			buf_snprintf(obuf, last, "%s %s ",PAR(p)->name, buf);
		}
	}
	return;
}


static int snprint_item(char **obuf, char *last, GcxPar p, int level)
{
	int pp = p;

//	d3_printf("snprint item %d obuf: %x\n", p, (unsigned) *obuf);

	if(PAR_TYPE(p) == PAR_TREE) {
//		sprintf(npath, "%s.%s", path, PAR(p)->name);
//		fprintf(fp, "# %s\n%s.\n\n", PAR(p)->comment, npath);
		if (level == 0) {
			buf_snprintf(obuf, last, "( ");
		} else {
			buf_snprintf(obuf, last, "%s ( ", PAR(p)->name);
		}
		p = PAR(pp)->child;
/* output the leaves first */
		while (p != PAR_NULL) {
			if (PAR_TYPE(p) != PAR_TREE) {
				snprint_item(obuf, last, p, level + 1);
			}
			p = PAR(p)->next;
		}
/* then the subtrees */
		p = PAR(pp)->child;
		while (p != PAR_NULL) {
			if (PAR_TYPE(p) == PAR_TREE) {
				snprint_item(obuf, last, p, level + 1);
			}
			p = PAR(p)->next;
		}
//		fprintf(fp, "\n");
		buf_snprintf(obuf, last, ") ");
	} else {
		snprint_par_leaf(obuf, last, p, level);
	}
	return 0;
}


/* string_print params.
 * if p is PAR_NULL, save all the params, otherwise just the 
 * given param and it's children
 * returns a negative error code, or 0 for success
 */
void snprint_params(char *obuf, int sz, GcxPar p)
{
	char *last;
	int ret;

	last = obuf + sz - 1;
	if (p != PAR_NULL) {
		snprint_item(&obuf, last, p, 0);
		buf_snprintf(&obuf, last, "\n");
	} else {
		p = PAR_FIRST;
		buf_snprintf(&obuf, last, "( ");
		while (p != PAR_NULL) {
			ret = snprint_item(&obuf, last, p, 1);
			p = PAR(p)->next;
		}
		buf_snprintf(&obuf, last, ")\n");
	}
}


/* definitions for the lexical analyser */
typedef enum {
	T_START,
	T_STRING,
	T_WORD,
} ParLexState;
static inline int is_string_delim(char c) {return (c == '\'') || (c=='"');}
static inline int is_name_char(char c) {return (isalnum(c)) || (c=='_') || (c == '-');}
/* simple lexical scanner for param file reading 
 * returns the token type 
 * updates the start and end pointers, which 
 * give the first and last char of the token 
 * Also updates the text pointer to
 * the next char the reader will look at. 
 * The scanner never goes back, so the text behind pos
 * can be altered at will.
 * when called with text=NULL, it resets the lex state
 */
int next_token(char **text, char **start, char **end)
{
	static ParLexState state = T_START;
	static char string_delim;
	char c;

	if (text == NULL) {
		state = T_START;
		return TOK_EOL;
	}

	while (**text != 0) {
		c = **text; 
		switch(state) {
		case T_START:
			if (isspace(c))
				break;
			if (c == '.') {
				*start = *text;
				*end = *text;
				(*text) ++;
				return TOK_DOT;
			}
			if (c == '#') 
				return TOK_EOL;
			if (is_string_delim(c)) {
				string_delim = c;
				*start = *text;
				state = T_STRING;
				break;
			}
			if (is_name_char(c)) {
				*start = *text;
				state = T_WORD;
				break;
			}
			*start = *text;
			*end = *text;
			(*text) ++;
			return TOK_PUNCT;
		case T_STRING:
			if (c == string_delim) {
				state = T_START;
				(*text) ++;
				return TOK_STRING;
			}
			*end = *text;
			break;
		case T_WORD:
			if (!is_name_char(c)) {
				state = T_START;
//				(*text) ++;
				return TOK_WORD;
			}
			*end = *text;
			break;
		}
		(*text) ++;
	}
	state = T_START;
	return TOK_EOL;
}

/*
 * try to assign the given value string to par p
 */
static int parse_token_val(GcxPar p, char *start)
{
	int ret;
//	d3_printf("par %d gets value %s\n", p, start);
	ret = try_update_par_value(p, start);
	if (ret < 0) {
		err_printf("value %s for par %d (%s) doesn't parse\n",
			   start, p, PAR(p)->name);
		return ret;
	} else {
		if (PAR_TYPE(p) == PAR_STRING) {
			change_par_default_string(p, PAR(p)->val.s);
		} else {
			memcpy(&(PAR(p)->defval), &(PAR(p)->val), sizeof(union pval));
		}
		PAR_FLAGS(p) |= PAR_FROM_RC;
	}
	return 0;
}


/* lookup a node by name within the peers of p */
static GcxPar node_lookup(GcxPar p, char *name, int len)
{
	while (p != PAR_NULL) {
//		d3_printf("lookup [%s] %d %s\n", PAR(p)->name, len, name);
		if (name_matches(PAR(p)->name, name, len)) {
//			d3_printf("match\n");
			return p;
		}
		p = PAR(p)->next;
	}
	return PAR_NULL;
}


typedef enum {
	P_START,
	P_NLINE,
	P_DOT,
	P_REL_TREE,
	P_ABS_TREE,
	P_LEAF,
	P_NAME,
	P_AFTER_NAME,
} ParParseStat;
/*
 * scan a branch of the param tree
 * if p is PAR_NULL, the root of the file is placed
 * at the root of the par tree, else it is assumed to be p
 */
int fscan_params(FILE *fp, GcxPar root)
{
	ParParseStat state = P_START;
	char *line=NULL, *text, *start, *end;
	int ret=0, len, token;
	GcxPar pp = PAR_NULL; /* tree we're reading in */
	GcxPar p = PAR_NULL; /* current item we read in */

	if (root == PAR_NULL)
		root = PAR_FIRST;
	else if (PAR(root)->child == PAR_NULL) {
		err_printf("%d - not a tree\n", root);
		return -1;
	} else 
		root = PAR(root)->child;

	while (ret >= 0) {
//		d3_printf("parse state is %d\n", state);
		switch(state) {
		case P_START:
			line = NULL;
			len = 0;
			pp = root;
			p = PAR_NULL;
			ret = 0;
			state = P_NLINE;
			break;
		case P_NLINE:
			ret = getline (&line, &len, fp);
			if (ret < 0) 
				break;
			text = line;
			token = next_token(&text, &start, &end);
			switch(token) {
			case TOK_EOL:
				ret = 0;
				break;
			case TOK_WORD: /* start of a relative path */
				p = node_lookup(pp, start, end-start);
				if (p == PAR_NULL) { 
					*(end+1)=0;
					err_printf("fscan_params: unknown name %s under node %d\n",
						   start, PAR(pp)->parent);
					state = P_NLINE;
				} else {
					if (PAR_TYPE(p) != PAR_TREE) {
						state = P_LEAF;
					} else {
						state = P_REL_TREE;
					}
				}
				break;
			case TOK_DOT: /* start of an absolute path */
				pp = root;
				state = P_ABS_TREE;
				break;
			default:
				*(end+1)=0;
				err_printf("fscan_params: unexpected token %s at start of line\n",
					   start);
				state = P_NLINE;
				break;
			}
			break;
		case P_REL_TREE:
			token = next_token(&text, &start, &end);
			if (token == TOK_DOT)
				break;
			if (token != TOK_WORD) {
				*(end+1)=0;
				err_printf("fscan_params: unexpected token %s at start of line\n",
					   start);
				state = P_NLINE;
				break;
			}
			p = node_lookup(p, start, end-start);
			if (p == PAR_NULL) { 
				*(end+1)=0;
				err_printf("fscan_params: unknown name %s under node %d\n",
					   start, PAR(p)->parent);
				state = P_NLINE;
				break;
			} else {
				if (PAR_TYPE(p) != PAR_TREE) {
					state = P_LEAF;
				} 
			}
			break;
		case P_ABS_TREE:
			token = next_token(&text, &start, &end);
			if (token == TOK_DOT)
				break;
			if (token == TOK_EOL) {
				state = P_NLINE;
				break;
			}
			if (token == TOK_WORD) {
				p = node_lookup(pp, start, end-start);
				if (p == PAR_NULL) { 
					*(end+1)=0;
					err_printf("fscan_params: unknown name %s under node %d\n",
						   start, PAR(pp)->parent);
				} else {
					if (PAR_TYPE(p) != PAR_TREE) {
						state = P_LEAF;
					} else {
						pp = PAR(p)->child;
					}
				}
			} else {
				*(end+1)=0;
				err_printf("fscan_params: unexpected token %s\n",
					   start);
				state = P_NLINE;
				break;
			}
			break;
		case P_LEAF: 
			token = next_token(&text, &start, &end);
			if (token == TOK_DOT)
				break;
			if (token == TOK_WORD || token == TOK_STRING) {
				*(end+1) = 0;
				parse_token_val(p, start);
				state = P_NLINE;
				break;
			}
			*(end+1) = 0;
			err_printf("fscan_params: unexpected token %s instead of par value\n",
				   start);
			break;
		default:
			break;
		}
	}
	if (line != NULL)
		free(line);
	return 0;
}

/* check that the name staring at name and len chars long matches the
 * string str. return 1 for a match */
int name_matches(char *str, char *name, int len)
{
	while (*str != 0 && len > 0) {
		if (*str == *name) {
			str ++;
			name ++;
			len --;
		} else {
			return 0;
		}
	}
	return !(*str || len);
}

/* lookup a node by it's pathname. The pathname 
 * is assumed to start at root level. More is updated with the start 
 * of the first token after the name. 
 */ 
GcxPar par_lookup(char *path, char **more)
{
	ParParseStat state = P_START;
	char *text, *start, *end;
	int token;
	GcxPar p, pp;

	text = path;
	*more = text;
	pp = PAR_FIRST;
	p = PAR_NULL;
	next_token(NULL, NULL, NULL);

	if (path == NULL)
		return PAR_NULL;

	while (*text != 0) {
		token = next_token(&text, &start, &end);
		*more = text;
//		d3_printf("parse token: %d [%s] state: %d text: %s\n", token, start, state, text);
		switch (state) {
		case P_START:
			if (token == TOK_DOT) {
				state = P_NAME;
//				p = pp;
				break;
			}
			state = P_NAME;
		/* fallthrough */
		case P_NAME:
			if (token != TOK_WORD)
				return p;
			p = node_lookup(pp, start, end-start+1);
//			d3_printf("lookup of %s returns %d\n", start, p);
			if (p == PAR_NULL)
				return PAR_NULL;
			if (PAR(p)->child != PAR_NULL)
				pp = PAR(p)->child;
			else
				return p;
			state = P_AFTER_NAME;
			break;
		case P_AFTER_NAME:
			if (token != TOK_DOT)
				return p;
			state = P_NAME;
			break;
		default:
			break;
		}
	}
	return PAR_NULL;
}

/* try to set a par to the given value string 
 * return 0 if par was set, -1 in case of error
 */
int set_par(GcxPar p, char *val)
{
	char *text, *start, *end;
	int token;

	text = val;
	next_token(NULL, NULL, NULL);
	token = next_token(&text, &start, &end);
//	d3_printf("parse token: %d [%s] text: %s\n", token, start, text);
	if (token != TOK_WORD && token != TOK_STRING) 
		return -1;
	*(end+1) = 0;
	return parse_token_val(p, start);
}
