/* Functions that handle 'parameters' - global options that are 
 * saved/loaded from rc files
 *
 * parameters are created at compile-time. Their names are built
 * hierarchically, but accessed from the rest of the code
 * trough macros taking constant indices as arguments
 */

#ifndef _PARAMS_H_
#define _PARAMS_H_

/* lexical analyser tokens */
typedef enum {
	TOK_EOL,
	TOK_WORD,
	TOK_DOT,
	TOK_STRING,
	TOK_PUNCT,
} TokenType;

union pval {
	int i;
	unsigned int u;
	double d;
	char *s;
};

struct param {
	int flags; /* data type & flags of the parameter */
	char ** choices; /* list of param choices / limits */
	int next; /* this node's next sibling */
	int child; /* this node's first child (NULL for leaf nodes) */
	int parent; /* this node's parent (NULL for the root node) */
	char *name; /* name as visible in the rc file */
	char *comment; /* a comment describing the parameter */
	union pval val;
	union pval defval;
};

extern struct param ptable[];

typedef int GcxPar;

/* param types */
#define PAR_TYPE_MASK 0xff
#define PAR_INVALID 0
#define PAR_TREE 1
#define PAR_INTEGER 0x10
#define PAR_DOUBLE 0x30
#define PAR_STRING 0x40

/* the format field tells us how we should print and read the parameter */
#define PAR_FORMAT_FM 0x0f00

#define FMT_NAT 0
#define FMT_RA 0x100 /* dms r.a mode for double */
#define FMT_DEC 0x200 /* dms r.a mode for double */
#define FMT_BOOL 0x300 /* boolean (for int) */
#define FMT_HEX 0x400 /* hex (for int) */
#define FMT_COLOR 0x500 /* color (r,g,b) format */
#define FMT_OPTION 0x600 /* one of several options (for int) 
			  * the options are kept in the NULL-terminated
			  * choices string list, the value being each option's
			  * position */

#define PAR_PREC_FM 0xf000 /* precision field for double */
#define PAR_PREC_FS 12 

#define PREC_1 (1 << PAR_PREC_FS)
#define PREC_2 (2 << PAR_PREC_FS)
#define PREC_4 (4 << PAR_PREC_FS)
#define PREC_8 (8 << PAR_PREC_FS)

/* flags */
#define PAR_FROM_RC 0x10000 /* read from a rc file */
#define PAR_USER 0x20000 /* changed by user */
#define PAR_TO_SAVE 0x40000 /* to be saved */
#define PAR_WREN 0x80000 /* par can be written */

/* macros to acces the par values */
#define P_INT(p) (ptable[p].val.i) 
#define P_UINT(p) (ptable[p].val.u)
#define P_DBL(p) (ptable[p].val.d)
#define P_STR(p) ((ptable[p].val.s))
#define PAR(p) ((p)== 0 ? NULL : &(ptable[p]))
#define PAR_TYPE(p) (ptable[p].flags & PAR_TYPE_MASK)
#define PAR_FORMAT(p) (ptable[p].flags & PAR_FORMAT_FM)
#define PAR_PREC(p) (ptable[p].flags & PAR_PREC_FM)
#define PAR_FLAGS(p) (ptable[p].flags)
/* parameter indices */

/* trees */
#define PAR_NULL 0
#define PAR_FIRST 1
#define MAIN_CAM 1
#define COOLER 2 
#define M_CAM_INFO 3
#define M_CAM_EXP 4
#define M_CAM_STAT 5

#define TRACK_CAM 40
#define T_CAM_INFO 41
#define T_CAM_EXP 42
#define T_CAM_STAT 43

#define TEMP_SETPOINT 6
#define TEMP_CURRENT 7
#define COOLING_POWER 8

#define M_CAM_NAME 10
#define M_ACTIVE_PIX 11
#define M_ACTIVE_LINES 12
#define M_DUMMY_PIX_START 13
#define M_DUMMY_PIX_END 14
#define M_DUMMY_LINES_START 15
#define M_DUMMY_LINES_END 16
#define M_PIX_X_SIZE 17
#define M_PIX_Y_SIZE 18
#define M_MIN_EXP 19
#define M_MAX_EXP 20

#define M_WIDTH 25
#define M_HEIGHT 26
#define M_BIN_X 27
#define M_BIN_Y 28
#define M_SKIP_X 29
#define M_SKIP_Y 30
#define M_EXPTIME 31
#define M_JDATE 37

#define M_RDNOISE 32
#define M_SCALE 33

#define M_STATUS 34
#define M_EXP_LEFT 35
#define M_LINES_READ 36

#define T_CAT_NAME 45
#define T_ACTIVE_PIX 46
#define T_ACTIVE_LINES 47
#define T_DUMMY_PIX_START 48
#define T_DUMMY_PIX_END 49
#define T_DUMMY_LINES_START 50
#define T_DUMMY_LINES_END 51
#define T_PIX_X_SIZE 52
#define T_PIX_Y_SIZE 53
#define T_MIN_EXP 54
#define T_MAX_EXP 55

#define T_WIDTH 56
#define T_HEIGHT 57
#define T_BIN_X 58
#define T_BIN_Y 59
#define T_SKIP_X 60
#define T_SKIP_Y 61
#define T_EXPTIME 62
#define T_RDNOISE 63
#define T_SCALE 64

#define T_STATUS 65
#define T_EXP_LEFT 66
#define T_LINES_READ 67

#define PAR_TABLE_SIZE 100

/* function prototypes */
/* params.c */
extern void init_ptable(void);
char *status_string(GcxPar p);
void make_value_string(GcxPar p, char *c);
int try_parse_double(GcxPar p, char *text);
int try_parse_int(GcxPar p, char *text);
int try_update_par_value(GcxPar p, char *text);
void change_par_string(GcxPar p, char *text);
int or_child_type(GcxPar p, int type);
int fscan_params(FILE *fp, GcxPar root);
int name_matches(char *str, char *name, int len);
GcxPar par_lookup(char *path, char **more);
void snprint_params(char *obuf, int sz, GcxPar p);
int set_par(GcxPar p, char *val);
int next_token(char **text, char **start, char **end);

#endif
